import os
from typing import List

import nrrd
import numpy as np

from supervisely import logger
from supervisely.io.fs import file_exists
from supervisely.volume.volume import read_dicom_serie_volume_np


def dcm_to_nrrd(id: str, paths: List[str]) -> str:
    """Convert DICOM series to NRRD format."""

    volume_np, volume_meta = read_dicom_serie_volume_np(paths)
    logger.info(f"PatientID and PatientName field for {id} will be anonymized.")
    serie_dir = os.path.dirname(paths[0])
    parent_dir = os.path.dirname(serie_dir)
    nrrd_path = os.path.join(parent_dir, f"{id}.nrrd")
    i = 0
    while file_exists(nrrd_path):
        i += 1
        nrrd_path = os.path.join(parent_dir, f"{id}_{i}.nrrd")

    nrrd_header = {
        "space": volume_meta.get("ACS", "RAS"),
    }
    if volume_meta.get("origin", None) is not None:
        nrrd_header["space origin"] = volume_meta.get("origin")
    directions = volume_meta.get("directions", None)
    spacing = volume_meta.get("spacing", None)
    if directions is not None and spacing is not None:
        directions = np.array(directions).reshape(3, 3)
        directions *= spacing
        nrrd_header["space directions"] = directions

    nrrd.write(nrrd_path, volume_np, nrrd_header)

    return nrrd_path, volume_meta

def convert_to_monochrome2(dcm_path: str):
    import pydicom

    is_modified = False

    try:
        dcm = pydicom.dcmread(dcm_path)
    except Exception as e:
        logger.warn("Failed to read DICOM file: " + str(e))
        return

    try:
        if dcm.file_meta.TransferSyntaxUID.is_compressed:
            dcm.decompress()
            is_modified = True
    except Exception as e:
        logger.warn("Failed to decompress DICOM file: " + str(e))
        return

    if getattr(dcm, "PhotometricInterpretation", None) == "YBR_FULL_422":
        # * Convert dicom to monochrome
        if len(dcm.pixel_array.shape) == 4 and dcm.pixel_array.shape[-1] == 3:
            monochrome = dcm.pixel_array[..., 0].astype(np.uint8)
        else:
            logger.warn("Unexpected shape for YBR_FULL_422 data: " + str(dcm.pixel_array.shape))

        try:
            dcm.SamplesPerPixel = 1
            dcm.PhotometricInterpretation = "MONOCHROME2"
            dcm.PlanarConfiguration = 0
            if len(monochrome.shape) == 3:
                dcm.NumberOfFrames = str(monochrome.shape[0])
                dcm.Rows, dcm.Columns = monochrome.shape[1:3]
            dcm.PixelData = monochrome.tobytes()
        except AttributeError as ae:
            logger.error(f"Error occurred while converting dicom to monochrome: {ae}")

        logger.info("Rewriting DICOM file with MONOCHROME2 photometric interpretation.")
        is_modified = True

    try:
        if is_modified:
            dcm.save_as(dcm_path)
    except Exception as e:
        logger.warn("Failed to save DICOM file: " + str(e))



# coding: utf-8
"""collection with :class:`Tag<supervisely.annotation.tag.Tag>` instances"""

# docs
from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional

from supervisely.annotation.tag import Tag
from supervisely.annotation.tag_meta import TagMeta
from supervisely.annotation.tag_meta_collection import TagMetaCollection
from supervisely.collection.key_indexed_collection import MultiKeyIndexedCollection


class TagCollection(MultiKeyIndexedCollection):
    """
    Collection with :class:`Tag<supervisely.annotation.tag.Tag>` instances. :class:`TagCollection<TagCollection>` object is immutable.

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # Create TagMetas (see class TagMeta for additional information about creating TagMetas)
        meta_weather = sly.TagMeta('Weather', sly.TagValueType.ANY_STRING)

        seasons = ["Winter", "Spring", "Summer", "Autumn"]
        meta_season = sly.TagMeta('Season', sly.TagValueType.ONEOF_STRING, possible_values=seasons)

        # Create Tags
        tag_weather = sly.Tag(meta_weather, value="Sunny")
        tag_season = sly.Tag(meta_season, value="Spring")
        tags_arr = [tag_weather, tag_season]

        # Create TagCollection from Tags
        tags = sly.TagCollection(tags_arr)

        # Add item to TagCollection
        meta_potato = sly.TagMeta('potato', sly.TagValueType.NONE)
        tag_potato =sly.Tag(meta_potato)

        # Remember that TagCollection is immutable, and we need to assign new instance of TagCollection to a new variable
        tags = tags.add(tag_potato)

        # You can also add multiple items to collection
        meta_cabbage = sly.TagMeta('cabbage', sly.TagValueType.NONE)
        meta_carrot = sly.TagMeta('carrot', sly.TagValueType.NONE)
        meta_turnip = sly.TagMeta('turnip', sly.TagValueType.NONE)

        tag_cabbage = sly.Tag(meta_cabbage)
        tag_carrot = sly.Tag(meta_carrot)
        tag_turnip = sly.Tag(meta_turnip)

        additional_veggies = [tag_cabbage, tag_carrot, tag_turnip]

        tags = tags.add_items(additional_veggies)

        # Has key, checks if given key exist in collection
        tags.has_key("cabbage")
        # Output: True

        # Intersection, finds intersection of given list of instances with collection items
        meta_dog = sly.TagMeta('dog', sly.TagValueType.NONE)
        tag_dog = sly.Tag(meta_dog)

        meta_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
        tag_cat = sly.Tag(meta_cat)

        meta_turtle = sly.TagMeta('turtle', sly.TagValueType.NONE)
        tag_turtle = sly.Tag(meta_turtle)


        tags_animals = sly.TagCollection([tag_dog, tag_cat, tag_turtle])

        tags_intersections = tags.intersection(tags_animals)
        print(tags_intersections.to_json())
        # Output: []

        # Let's add the potato Tag from another collection and compare them again
        tags_animals = tags_animals.add(tag_potato)

        tags_intersections = tags.intersection(tags_animals)
        print(tags_intersections.to_json())
        # Output: [
        #     {
        #         "name":"potato"
        #     }
        # ]

        # Difference, finds difference between collection and given list of Tags or TagCollection
        meta_car = sly.TagMeta('car', sly.TagValueType.NONE)
        tag_car = sly.Tag(meta_car)

        meta_bicycle = sly.TagMeta('bicycle', sly.TagValueType.NONE)
        tag_bicycle = sly.Tag(meta_bicycle)

        tags_vehicles = sly.TagCollection([tag_car, tag_bicycle])

        meta_pedestrian = sly.TagMeta('pedestrian', sly.TagValueType.NONE)
        tag_pedestrian = sly.Tag(meta_pedestrian)

        meta_road = sly.TagMeta('road', sly.TagValueType.NONE)
        tag_road = sly.Tag(meta_road)

        difference = tags_vehicles.difference([tag_pedestrian, tag_road])
        print(difference.to_json())
        # Output: [
        #     {
        #         "name":"car"
        #     },
        #     {
        #         "name":"bicycle"
        #     }
        # ]

        # Merge, merges collection and given list of TagMetas
        tags_vehicles = sly.TagMetaCollection([tag_car, tag_bicycle])
        tags_merge = sly.TagMetaCollection([tag_pedestrian, tag_road])

        merged_collections = tags_vehicles.merge(tags_merge)
        print(merged_collections.to_json())
        # Output: [
        #     {
        #         "name":"pedestrian"
        #     },
        #     {
        #         "name":"road"
        #     },
        #     {
        #         "name":"car"
        #     },
        #     {
        #         "name":"bicycle"
        #     }
        # ]
    """

    item_type = Tag

    def to_json(self) -> List[Dict]:
        """
        Convert the TagCollection to a list of json dicts. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :return: List of dicts in json format
        :rtype: :class:`List[dict]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            meta_weather = sly.TagMeta('Weather', sly.TagValueType.ANY_STRING)

            seasons = ["Winter", "Spring", "Summer", "Autumn"]
            meta_season = sly.TagMeta('Season', sly.TagValueType.ONEOF_STRING, possible_values=seasons)

            tag_weather = sly.Tag(meta_weather, value="Sunny")
            tag_season = sly.Tag(meta_season, value="Spring")

            tags_arr = [tag_weather, tag_season]
            tags = sly.TagCollection(tags_arr)

            tags_json = tags.to_json()
            # Output: [
            #     {
            #         "name":"Weather",
            #         "value":"Sunny"
            #     },
            #     {
            #         "name":"Season",
            #         "value":"Spring"
            #     }
            # ]
        """
        return [tag.to_json() for tag in self]

    @classmethod
    def from_json(cls, data: List[Dict], tag_meta_collection: TagMetaCollection) -> TagCollection:
        """
        Convert a list with dicts in json format to TagCollection. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: List with dicts in json format.
        :type data: List[dict]
        :param tag_meta_collection: Input TagMetaCollection object.
        :type tag_meta_collection: TagMetaCollection
        :return: TagCollection object
        :rtype: :class:`TagCollection<TagCollection>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            # Initialize TagMetaCollection

            meta_weather = sly.TagMeta('Weather', sly.TagValueType.ANY_STRING)

            seasons = ["Winter", "Spring", "Summer", "Autumn"]
            meta_season = sly.TagMeta('Season', sly.TagValueType.ONEOF_STRING, possible_values=seasons)

            metas_arr = [meta_weather, meta_season]
            tag_metas = sly.TagMetaCollection(metas_arr)

            data = [
                {
                    "name":"Weather",
                    "value":"Sunny"
                },
                {
                    "name":"Season",
                    "value":"Spring"
                }
            ]

            tags = sly.TagCollection.from_json(data, tag_metas)
        """
        tags = [cls.item_type.from_json(tag_json, tag_meta_collection) for tag_json in data]
        return cls(tags)

    def __str__(self):
        return "Tags:\n" + super(TagCollection, self).__str__()

    @classmethod
    def from_api_response(
        cls,
        data: List[Dict],
        tag_meta_collection: TagMetaCollection,
        id_to_tagmeta: Optional[Dict[int, TagMeta]] = None,
    ) -> TagCollection:
        """
        Create a TagCollection object from API response data.

        :param data: API response data.
        :type data: List[Dict]
        :param tag_meta_collection: TagMetaCollection object
        :type tag_meta_collection: TagMetaCollection
        :param id_to_tagmeta: Mapping of tag IDs to tag metadata.
        :type id_to_tagmeta: Optional[Dict[int, TagMeta]]
        :return: TagCollection object.
        :rtype: TagCollection
        :Usage example:

        .. code-block:: python

            import supervisely as sly

            # You can connect to API directly
            address = 'https://app.supervisely.com/'
            token = 'Your Supervisely API Token'
            api = sly.Api(address, token)

            # Or you can use API from environment
            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_id = 17200
            image_id = 19369643
            project_meta_json = api.project.get_meta(project_id)
            project_meta = sly.ProjectMeta.from_json(project_meta_json)
            image_info = api.image.get_info_by_id(image_id)

            tag_collection = sly.TagCollection.from_api_response(
                image_info.tags, project_meta.tag_metas
            )
            print(tag_collection)

            # Output:
            # Tags:
            # +-------+--------------+-------+
            # |  Name |  Value type  | Value |
            # +-------+--------------+-------+
            # | Lemon | oneof_string |  big  |
            # |  Kiwi |     none     |  None |
            # +-------+--------------+-------+
        """

        if id_to_tagmeta is None:
            id_to_tagmeta = tag_meta_collection.get_id_mapping()
        tags = []
        for tag_json in data:
            tag_meta_id = tag_json["tagId"]
            tag_meta = id_to_tagmeta[tag_meta_id]
            tag_json["name"] = tag_meta.name
            tag = cls.item_type.from_json(tag_json, tag_meta_collection)
            tags.append(tag)
        return cls(tags)

    def __iter__(self) -> Iterator[Tag]:
        return next(self)



import os
from collections import defaultdict
from typing import Dict, Optional, Set, Union

from supervisely import (
    Annotation,
    ProjectMeta,
    TagApplicableTo,
    TagMeta,
    TagValueType,
    logger,
)
from supervisely.convert.base_converter import AvailableImageConverters
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.convert.image.pascal_voc import pascal_voc_helper
from supervisely.io.fs import (
    dir_exists,
    dirs_filter,
    file_exists,
    get_file_ext,
    get_file_name,
    list_files_recursively,
)
from supervisely.project.project_settings import LabelingInterface


class PascalVOCConverter(ImageConverter):
    class Item(ImageConverter.Item):
        def __init__(
            self,
            *args,
            **kwargs,
        ):
            super().__init__(*args, **kwargs)
            self._segm_path = None
            self._inst_path = None

        @property
        def segm_path(self) -> str:
            return self._segm_path

        @property
        def inst_path(self) -> str:
            return self._inst_path

        @segm_path.setter
        def segm_path(self, segm_path: str) -> None:
            self._segm_path = segm_path

        @inst_path.setter
        def inst_path(self, inst_path: str) -> None:
            self._inst_path = inst_path

    def __init__(
        self,
        input_data: str,
        labeling_interface: Optional[Union[LabelingInterface, str]],
        upload_as_links: bool,
        remote_files_map: Optional[Dict[str, str]] = None,
    ):
        super().__init__(input_data, labeling_interface, upload_as_links, remote_files_map)

        self.color2class_name: Optional[Dict[str, str]] = None
        self.with_instances: bool = False
        self._imgs_dir: Optional[str] = None
        self._segm_dir: Optional[str] = None
        self._inst_dir: Optional[str] = None
        self._bbox_classes_map: Dict[str, str] = {}

    def __str__(self) -> str:
        return AvailableImageConverters.PASCAL_VOC

    @property
    def ann_ext(self) -> str:
        return ".xml"

    def validate_ann_file(self, ann_path: str, meta: ProjectMeta) -> bool:
        # TODO: implement annotation validation
        pass

    def validate_format(self) -> bool:
        xml_files_exist = self._scan_for_xml_files()
        possible_pascal_voc_dir = self._scan_for_segm_dirs()
        if not possible_pascal_voc_dir and not xml_files_exist:
            return False
        if not possible_pascal_voc_dir:
            self._imgs_dir = self._input_data
            possible_pascal_voc_dir = self._input_data

        self._meta = self._generate_meta(possible_pascal_voc_dir)
        detected_ann_cnt = self._create_items(possible_pascal_voc_dir)
        return detected_ann_cnt > 0

    def _scan_for_xml_files(self) -> bool:
        for _, _, file_names in os.walk(self._input_data):
            for filename in file_names:
                if get_file_ext(filename) == self.ann_ext:
                    return True
        return False

    def _scan_for_segm_dirs(self) -> Optional[str]:
        def check_function(dir_path: str) -> bool:
            possible_image_dirs = ["JPEGImages", "Images", "images", "imgs", "img"]
            possible_segm_dirs = [
                "SegmentationClass",
                "segmentation",
                "segmentations",
                "Segmentation",
                "Segmentations",
                "masks",
                "segm",
            ]
            possible_inst_dirs = ["SegmentationObject", "instances", "objects"]

            if not any([dir_exists(os.path.join(dir_path, p)) for p in possible_image_dirs]):
                return False
            if not any([dir_exists(os.path.join(dir_path, p)) for p in possible_segm_dirs]):
                return False
            for d in possible_image_dirs:
                if dir_exists(os.path.join(dir_path, d)):
                    self._imgs_dir = os.path.join(dir_path, d)
                    break
            for d in possible_segm_dirs:
                if dir_exists(os.path.join(dir_path, d)):
                    self._segm_dir = os.path.join(dir_path, d)
                    break
            for d in possible_inst_dirs:
                if dir_exists(os.path.join(dir_path, d)):
                    self._inst_dir = os.path.join(dir_path, d)
                    break

            return self._imgs_dir is not None and self._segm_dir is not None

        possible_pascal_voc_dir = [d for d in dirs_filter(self._input_data, check_function)]
        if len(possible_pascal_voc_dir) > 1:
            logger.warn("Multiple Pascal VOC directories not supported")
            return
        elif len(possible_pascal_voc_dir) == 0:
            return
        else:
            return possible_pascal_voc_dir[0]

    def _generate_meta(self, possible_pascal_voc_dir: str) -> ProjectMeta:
        colors_file = os.path.join(possible_pascal_voc_dir, "colors.txt")
        obj_classes, self.color2class_name = pascal_voc_helper.read_colors(colors_file)
        return ProjectMeta(obj_classes=obj_classes)

    def _create_items(self, possible_pascal_voc_dir: str) -> int:
        existing_cls_names = set([cls.name for cls in self._meta.obj_classes])
        tags_to_values = defaultdict(set)
        detected_ann_cnt = 0

        images_list = list_files_recursively(self._imgs_dir, valid_extensions=self.allowed_exts)
        img_ann_map = {
            get_file_name(path): path
            for path in list_files_recursively(possible_pascal_voc_dir, [self.ann_ext])
        }

        self._items = []
        for image_path in images_list:
            item = self.Item(image_path)
            item_name_noext = get_file_name(item.name)
            item = self._scan_for_item_segm_paths(item, item_name_noext)
            ann_path = img_ann_map.get(item_name_noext) or img_ann_map.get(item.name)
            item = self._scan_for_item_ann_path_and_update_meta(
                item, ann_path, existing_cls_names, tags_to_values
            )

            if item.ann_data or item.segm_path:
                detected_ann_cnt += 1
            self._items.append(item)
        self._meta = self._update_meta_with_tags(tags_to_values)
        return detected_ann_cnt

    def _update_meta_with_tags(self, tags_to_values: Dict[str, Set[str]]) -> ProjectMeta:
        meta = self._meta
        object_class_names = set(meta.obj_classes.keys())
        for tag_name, values in tags_to_values.items():
            tag_meta = meta.get_tag_meta(tag_name)
            if tag_meta is not None:
                continue
            if tag_name in pascal_voc_helper.DEFAULT_SUBCLASSES:
                if values.difference({"0", "1"}):
                    logger.warning(
                        f"Tag '{tag_name}' has non-binary values.", extra={"values": values}
                    )
                tag_meta = TagMeta(tag_name, TagValueType.NONE)
            elif tag_name in object_class_names:
                tag_meta = TagMeta(
                    tag_name,
                    TagValueType.ONEOF_STRING,
                    possible_values=list(values),
                    applicable_to=TagApplicableTo.OBJECTS_ONLY,
                    applicable_classes=[tag_name],
                )
            else:
                tag_meta = TagMeta(tag_name, TagValueType.ANY_STRING)
            meta = meta.add_tag_meta(tag_meta)
        return meta

    def _scan_for_item_segm_paths(self, item: Item, item_name_noext: str) -> Item:
        if self._segm_dir is not None:
            segm_path = os.path.join(self._segm_dir, f"{item_name_noext}.png")
            if file_exists(segm_path):
                item.segm_path = segm_path
        if self._inst_dir is not None:
            inst_path = os.path.join(self._inst_dir, f"{item_name_noext}.png")
            if file_exists(inst_path):
                item.inst_path = inst_path

        return item

    def _scan_for_item_ann_path_and_update_meta(
        self,
        item: Item,
        ann_path: Optional[str],
        existing_cls_names: Set[str],
        tags_to_values: Dict[str, Set[str]],
    ) -> Item:
        if ann_path is None:
            return item
        if not file_exists(ann_path):
            return item
        self._meta = pascal_voc_helper.update_meta_from_xml(
            ann_path, self._meta, existing_cls_names, self._bbox_classes_map, tags_to_values
        )
        item.ann_data = ann_path
        return item

    def to_supervisely(
        self,
        item: Item,
        meta: Optional[ProjectMeta] = None,
        renamed_classes: Optional[Dict[str, str]] = None,
        renamed_tags: Optional[Dict[str, str]] = None,
    ) -> Annotation:
        if meta is None:
            meta = self._meta

        try:
            item.set_shape()
            return pascal_voc_helper.get_ann(
                item,
                self.color2class_name,
                meta,
                self._bbox_classes_map,
                renamed_classes,
                renamed_tags,
            )
        except Exception as e:
            logger.warn(f"Failed to convert annotation: {repr(e)}")
            return item.create_empty_annotation()



from typing import Dict, Literal, Optional, Union

from supervisely.app import StateJson
from supervisely.app.widgets import Widget

SUPPORTED_TAG_WIDGET_TYPES = ["primary", "gray", "success", "warning", "danger"]


class ElementTag(Widget):
    """ElementTag widget in Supervisely is a widget that allows users to display elements tag in the UI.

    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/text-elements/elementtag>`_
        (including screenshots and examples).

    :param text: Tag text
    :type text: Optional[str]
    :param type: Tag type, one of: primary, gray, success, warning, danger
    :type type: Optional[Literal["primary", "gray", "success", "warning", "danger"]]
    :param hit: If True, tag will be highlighted
    :type hit: Optional[bool]
    :param color: Tag color
    :type color: Optional[str]
    :param widget_id: An identifier of the widget.
    :type widget_id: str, optional

    :Usage example:
    .. code-block:: python

            from supervisely.app.widgets import ElementTag

            element_tag = ElementTag(
                text="Tag",
                type="primary",
                hit=True,
                color="#20a0ff",
            )
    """

    class Routes:
        CLOSE = "tag_close_cb"

    def __init__(
        self,
        text: Optional[str] = "",
        type: Optional[Literal["primary", "gray", "success", "warning", "danger", None]] = None,
        hit: Optional[bool] = False,
        color: Optional[str] = "",
        widget_id: Optional[str] = None,
    ):
        self._text = text
        self._validate_type(type)
        self._type = type
        self._color = color
        self._hit = hit
        self._tag_close = False
        self._clicked_tag = None

        super().__init__(widget_id=widget_id, file_path=__file__)

    def _validate_type(self, value):
        if value is None:
            return
        if value not in SUPPORTED_TAG_WIDGET_TYPES:
            raise ValueError(
                "type = {!r} is unknown, should be one of {}".format(
                    value, SUPPORTED_TAG_WIDGET_TYPES
                )
            )

    def get_json_data(self) -> Dict:
        """ElementTag widget has no data, the method returns an empty dict."""
        return {}

    def get_json_state(self) -> Dict[str, Union[str, bool]]:
        """Returns dictionary with widget state.

        The dictionary contains the following fields:
            - text: Tag text
            - type: Tag type, one of: primary, gray, success, warning, danger
            - hit: If True, tag will be highlighted
            - color: Tag color

        :return: dictionary with widget state
        :rtype: Dict[str, Union[str, bool]]
        """
        return {
            "text": self._text,
            "type": self._type,
            "hit": self._hit,
            "color": self._color,
        }

    @property
    def text(self) -> str:
        """Returns current tag text

        :return: current tag text
        :rtype: str
        """
        return self._text

    @property
    def type(self) -> Literal["primary", "gray", "success", "warning", "danger"]:
        """Returns current tag type

        :return: current tag type
        :rtype: Literal["primary", "gray", "success", "warning", "danger"]
        """
        return self._type

    @property
    def hit(self) -> bool:
        """Returns True if tag is highlighted, False otherwise

        :return: True if tag is highlighted, False otherwise
        :rtype: bool
        """
        return self._hit

    @property
    def color(self) -> str:
        """Returns current tag color.

        :return: current tag color
        :rtype: str
        """
        return self._color

    def set_text(self, value: str) -> None:
        """Sets current tag text.

        :param value: current tag text
        :type value: str
        """
        StateJson()[self.widget_id]["text"] = value
        StateJson().send_changes()

    def get_text(self) -> str:
        """Returns current tag text.

        :return: current tag text
        :rtype: str
        """
        return StateJson()[self.widget_id]["text"]

    def set_type(self, value: Literal["primary", "gray", "success", "warning", "danger"]) -> None:
        """Sets current tag type.

        :param value: current tag type
        :type value: Literal["primary", "gray", "success", "warning", "danger"]
        """
        self._validate_type(value)
        StateJson()[self.widget_id]["type"] = value
        StateJson().send_changes()

    def get_type(self) -> Literal["primary", "gray", "success", "warning", "danger"]:
        """Returns current tag type.

        :return: current tag type
        :rtype: Literal["primary", "gray", "success", "warning", "danger"]
        """
        return StateJson()[self.widget_id]["type"]

    def set_hit(self, value: bool) -> None:
        """Sets current tag highlight.

        :param value: current tag highlight
        :type value: bool
        """
        StateJson()[self.widget_id]["hit"] = value
        StateJson().send_changes()

    def get_hit(self) -> bool:
        """Returns True if tag is highlighted, False otherwise.

        :return: True if tag is highlighted, False otherwise
        :rtype: bool
        """
        return StateJson()[self.widget_id]["hit"]

    def set_color(self, value: str) -> None:
        """Sets current tag color.

        :param value: current tag color
        :type value: str
        """
        StateJson()[self.widget_id]["color"] = value
        StateJson().send_changes()

    def get_color(self) -> str:
        """Returns current tag color.

        :return: current tag color
        :rtype: str
        """
        return StateJson()[self.widget_id]["color"]



import json
import shutil
import tempfile
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from enum import Enum
from logging import Logger
from pathlib import Path
from threading import Lock, Thread
from typing import Any, BinaryIO, Callable, Generator, List, Optional, Tuple, Union

import cv2
import numpy as np
from cacheout import Cache as CacheOut
from cachetools import Cache, LRUCache, TTLCache
from fastapi import BackgroundTasks, FastAPI, Form, Request, UploadFile

import supervisely as sly
import supervisely.io.env as env
from supervisely._utils import batched
from supervisely.io.fs import list_files, silent_remove
from supervisely.video.video import VideoFrameReader


class PersistentImageLRUCache(LRUCache):
    __marker = object()

    def __init__(self, maxsize, filepath: Path, getsizeof=None):
        super().__init__(maxsize)
        self._base_dir = filepath

    def __getitem__(self, key: Any) -> Any:
        filepath = super(PersistentImageLRUCache, self).__getitem__(key)
        return sly.image.read(str(filepath))

    def __setitem__(self, key: Any, value: Any) -> None:
        if not self._base_dir.exists():
            self._base_dir.mkdir()

        filepath = self._base_dir / f"{str(key)}.png"
        super(PersistentImageLRUCache, self).__setitem__(key, filepath)

        if filepath.exists():
            sly.logger.debug(f"Rewrite image {str(filepath)}")
        sly.image.write(str(filepath), value)

    def pop(self, key, default=__marker):
        if key in self:
            filepath = self._base_dir / f"{str(key)}.png"
            value = self[key]
            del self[key]
            silent_remove(filepath)
            sly.logger.debug(f"Remove {filepath} frame")
        elif default is self.__marker:
            raise KeyError(key)
        else:
            value = default
        return value

    def clear(self, rm_base_folder=True) -> None:
        while self.currsize > 0:
            self.popitem()
        if rm_base_folder:
            shutil.rmtree(self._base_dir)


class PersistentImageTTLCache(TTLCache):
    def __init__(self, maxsize: int, ttl: int, filepath: Path):
        super().__init__(maxsize, ttl)
        self._base_dir = filepath

    def pop(self, *args, **kwargs):
        try:
            super().pop(*args, **kwargs)
        except Exception:
            sly.logger.warn("Cache data corrupted. Cleaning the cache...", exc_info=True)

            def _delitem(self, key):
                try:
                    size = self._Cache__size.pop(key)
                except:
                    size = 0
                self._Cache__data.pop(key, None)
                self._Cache__currsize -= size

            shutil.rmtree(self._base_dir, ignore_errors=True)
            for key in self.keys():
                try:
                    super().__delitem__(key, cache_delitem=_delitem)
                except:
                    pass

    def __delitem__(self, key: Any) -> None:
        self.__del_file(key)
        return super().__delitem__(key)

    def __del_file(self, key: str):
        cache_getitem = Cache.__getitem__
        filepath = cache_getitem(self, key)
        try:
            silent_remove(filepath)
        except TypeError:
            pass
        except:
            sly.logger.debug(f"File {filepath} was not deleted from cache", exc_info=True)
        else:
            sly.logger.debug(f"File {filepath} was deleted from cache")

    def __update_timer(self, key):
        try:
            # pylint: disable=no-member
            link = self._TTLCache__getlink(key)
            # pylint: disable=no-member
            if hasattr(link, "expire"):
                link.expire = self.timer() + self._TTLCache__ttl
            else:
                link.expires = self.timer() + self._TTLCache__ttl
        except KeyError:
            return

    def __getitem__(self, key: Any) -> Any:
        self.__update_timer(key)
        return super().__getitem__(key)

    def __get_keys(self):
        # pylint: disable=no-member
        return self._TTLCache__links.keys()

    def expire(self, time=None):
        """Remove expired items from the cache."""
        # pylint: disable=no-member
        existing_items = self._Cache__data.copy()
        super().expire(time)
        deleted = set(existing_items.keys()).difference(self.__get_keys())
        if len(deleted) > 0:
            for key in deleted:
                try:
                    silent_remove(existing_items[key])
                except TypeError:
                    pass
            sly.logger.debug(f"Deleted keys: {deleted}")
            sly.logger.trace(f"Existing files: {list_files(str(self._base_dir))}")

    def clear(self, rm_base_folder=True) -> None:
        while self.currsize > 0:
            self.popitem()
        if rm_base_folder:
            shutil.rmtree(self._base_dir)

    def save_image(self, key, image: Union[np.ndarray, BinaryIO, bytes], ext=".png") -> None:
        if not self._base_dir.exists():
            self._base_dir.mkdir()

        if ext is None or ext == "":
            ext = ".png"

        filepath = self._base_dir / Path(key).with_suffix(ext)
        self[key] = filepath

        if filepath.exists():
            sly.logger.debug(f"Rewrite image {str(filepath)}")
        if isinstance(image, np.ndarray):
            sly.image.write(str(filepath), image)
        elif isinstance(image, bytes):
            with open(filepath, "wb") as f:
                f.write(image)
        else:
            with open(filepath, "wb") as f:
                shutil.copyfileobj(image, f)

    def get_image_path(self, key: Any) -> Path:
        return self[key]

    def get_image(self, key: Any):
        return sly.image.read(str(self[key]))

    def save_video(self, key: Any, source: Union[str, Path, BinaryIO]) -> None:
        ext = ""
        if isinstance(source, Path):
            ext = source.suffix
        elif isinstance(source, str):
            ext = Path(source).suffix
        video_path = self._base_dir / f"video_{key}{ext}"
        self[key] = video_path

        if isinstance(source, (str, Path)):
            if str(source) != str(video_path):
                shutil.move(source, str(video_path))
        else:
            with open(video_path, "wb") as f:
                shutil.copyfileobj(source, f)
        sly.logger.debug(f"Video #{key} saved to {video_path}", extra={"video_id": key})

    def get_video_path(self, key: Any) -> Path:
        return self[key]

    def save_project_meta(self, key, value):
        self[key] = value

    def get_project_meta(self, project_meta_name):
        return self[project_meta_name]

    def copy_to(self, name, path):
        shutil.copyfile(str(self[name]), path)


class InferenceImageCache:
    class _LoadType(Enum):
        ImageId: str = "IMAGE"
        ImageHash: str = "HASH"
        Frame: str = "FRAME"
        Video: str = "VIDEO"

    def __init__(
        self,
        maxsize: int,
        ttl: int,
        is_persistent: bool = True,
        base_folder: str = env.smart_cache_container_dir(),
        log_progress: bool = False,
    ) -> None:
        self.is_persistent = is_persistent
        self._maxsize = maxsize
        self._ttl = ttl
        self._lock = Lock()
        self._load_queue = CacheOut(ttl=10 * 60)
        self.log_progress = log_progress
        self._download_executor = ThreadPoolExecutor(max_workers=5)

        if is_persistent:
            self._data_dir = Path(base_folder)
            self._data_dir.mkdir(parents=True, exist_ok=True)
            self._cache = PersistentImageTTLCache(maxsize, ttl, self._data_dir)
        else:
            self._cache = TTLCache(maxsize, ttl)

    def clear_cache(self):
        with self._lock:
            self._cache.clear(False)

    def download_image(self, api: sly.Api, image_id: int, related: bool = False):
        name = self._image_name(image_id)
        self._wait_if_in_queue(name, api.logger)

        if name not in self._cache:
            self._load_queue.set(name, image_id)
            api.logger.debug(f"Add image #{image_id} to cache")
            if not related:
                img = api.image.download_np(image_id)
            else:
                img = api.pointcloud.download_related_image(image_id)
            self._add_to_cache(name, img)
            return img

        api.logger.debug(f"Get image #{image_id} from cache")
        return self._cache.get_image(name)

    def download_images(self, api: sly.Api, dataset_id: int, image_ids: List[int], **kwargs):
        return_images = kwargs.get("return_images", True)

        def load_generator(image_ids: List[int]):
            return api.image.download_nps_generator(dataset_id, image_ids)

        return self._download_many(
            image_ids,
            self._image_name,
            load_generator,
            api.logger,
            return_images,
        )

    def download_image_by_hash(self, api: sly.Api, img_hash: str) -> np.ndarray:
        image_key = self._image_name(img_hash)
        self._wait_if_in_queue(image_key, api.logger)

        if image_key not in self._cache:
            self._load_queue.set(image_key, img_hash)
            image = api.image.download_nps_by_hashes([img_hash])
            self._add_to_cache(image_key, image)
            return image
        return self._cache.get_image(image_key)

    def download_images_by_hashes(
        self, api: sly.Api, img_hashes: List[str], **kwargs
    ) -> List[np.ndarray]:
        return_images = kwargs.get("return_images", True)

        def load_generator(img_hashes: List[str]):
            return api.image.download_nps_by_hashes_generator(img_hashes)

        return self._download_many(
            img_hashes,
            self._image_name,
            load_generator,
            api.logger,
            return_images,
        )

    def _read_frames_from_cached_video_iter(self, video_id, frame_indexes):
        video_path = self._cache.get_video_path(video_id)
        with VideoFrameReader(video_path, frame_indexes) as reader:
            for frame in reader.iterate_frames():
                yield frame

    def _read_frames_from_cached_video(
        self, video_id: int, frame_indexes: List[int]
    ) -> List[np.ndarray]:
        return [
            frame for frame in self._read_frames_from_cached_video_iter(video_id, frame_indexes)
        ]

    def get_frame_from_cache(self, video_id: int, frame_index: int) -> np.ndarray:
        name = self._frame_name(video_id, frame_index)
        if isinstance(self._cache, PersistentImageTTLCache):
            if name in self._cache:
                return self._cache.get_image(name)
            return self._read_frames_from_cached_video(video_id, [frame_index])[0]
        frame = self._cache.get(name)
        if frame is None:
            raise KeyError(f"Frame {frame_index} not found in video {video_id}")

    def get_video_frames_count(self, key):
        """
        Returns number of frames in the video
        """
        if not isinstance(self._cache, PersistentImageTTLCache):
            raise ValueError("Video frames count can be obtained only for persistent cache")
        video_path = self._cache.get_video_path(key)
        return VideoFrameReader(video_path).frames_count()

    def get_video_frame_size(self, key):
        """
        Returns height and width of the video frame. (h, w)
        """
        if not isinstance(self._cache, PersistentImageTTLCache):
            raise ValueError("Video frame size can be obtained only for persistent cache")
        video_path = self._cache.get_video_path(key)
        return VideoFrameReader(video_path).frame_size()

    def get_video_fps(self, key):
        """
        Returns fps of the video
        """
        if not isinstance(self._cache, PersistentImageTTLCache):
            raise ValueError("Video fps can be obtained only for persistent cache")
        video_path = self._cache.get_video_path(key)
        return VideoFrameReader(video_path).fps()

    def get_frames_from_cache(self, video_id: int, frame_indexes: List[int]) -> List[np.ndarray]:
        if isinstance(self._cache, PersistentImageTTLCache) and video_id in self._cache:
            return self._read_frames_from_cached_video(video_id, frame_indexes)
        else:
            return [
                self.get_frame_from_cache(video_id, frame_index) for frame_index in frame_indexes
            ]

    def frames_loader(
        self, api: sly.Api, video_id: int, frame_indexes: List[int]
    ) -> Generator[np.ndarray, None, None]:
        if not isinstance(self._cache, PersistentImageTTLCache):
            for frame_index in frame_indexes:
                yield self.download_frame(api, video_id, frame_index)
            return
        self.run_cache_task_manually(api, None, video_id=video_id)
        for i, frame_index in enumerate(frame_indexes):
            if video_id in self._cache:
                break
            yield self.download_frame(api, video_id, frame_index)
        if i < len(frame_indexes):
            for frame in self._read_frames_from_cached_video_iter(video_id, frame_indexes[i:]):
                yield frame

    def download_frame(self, api: sly.Api, video_id: int, frame_index: int) -> np.ndarray:
        name = self._frame_name(video_id, frame_index)
        self._wait_if_in_queue(name, api.logger)

        if name not in self._cache:
            if video_id in self._cache:
                api.logger.debug(
                    f"Get frame #{frame_index} for video #{video_id} from cache (video file)"
                )
                try:
                    return self.get_frame_from_cache(video_id, frame_index)
                except:
                    sly.logger.warning(
                        f"Frame {frame_index} not found in video {video_id}", exc_info=True
                    )

            self._load_queue.set(name, (video_id, frame_index))
            frame = api.video.frame.download_np(video_id, frame_index)
            self._add_to_cache(name, frame)
            api.logger.debug(f"Add frame #{frame_index} for video #{video_id} to cache")
            return frame

        api.logger.debug(f"Get frame #{frame_index} for video #{video_id} from cache")
        return self._cache.get_image(name)

    def download_frames(
        self, api: sly.Api, video_id: int, frame_indexes: List[int], **kwargs
    ) -> List[np.ndarray]:
        return_images = kwargs.get("return_images", True)
        redownload_video = kwargs.get("redownload_video", False)
        progress_cb = kwargs.get("progress_cb", None)

        if video_id in self._cache:
            try:
                frames = self.get_frames_from_cache(video_id, frame_indexes)
                if progress_cb is not None:
                    progress_cb(len(frame_indexes))
                if return_images:
                    return frames
            except:
                sly.logger.warning(
                    f"Frames {frame_indexes} not found in video {video_id}", exc_info=True
                )
                self._download_executor.submit(
                    self.download_video,
                    api,
                    video_id,
                    **{**kwargs, "return_images": False},
                )
        elif redownload_video:
            self._download_executor.submit(
                self.download_video,
                api,
                video_id,
                **{**kwargs, "return_images": False},
            )

        def name_constuctor(frame_index: int):
            return self._frame_name(video_id, frame_index)

        def load_generator(frame_indexes: List[int]):
            return api.video.frame.download_nps_generator(video_id, frame_indexes)

        return self._download_many(
            frame_indexes,
            name_constuctor,
            load_generator,
            api.logger,
            return_images,
            progress_cb=progress_cb,
            video_id=video_id,
        )

    def add_video_to_cache_by_io(self, video_id: int, video_io: BinaryIO) -> None:
        if isinstance(self._cache, PersistentImageTTLCache):
            with self._lock:
                self._cache.save_video(video_id, source=video_io)

    def add_video_to_cache(self, video_id: int, source: Union[str, Path, BinaryIO]) -> None:
        """
        Adds video to cache.
        """
        if isinstance(self._cache, PersistentImageTTLCache):
            with self._lock:
                self._cache.save_video(video_id, source)
                self._load_queue.delete(video_id)
            sly.logger.debug(f"Video #{video_id} added to cache", extra={"video_id": video_id})
        else:
            tmp_source = None
            if not isinstance(source, (str, Path)):
                with tempfile.NamedTemporaryFile(delete=False) as f:
                    shutil.copyfileobj(source, f)
                    tmp_source = f.name
                    source = tmp_source
            try:
                with VideoFrameReader(source) as reader:
                    for frame_index, frame in enumerate(reader):
                        self.add_frame_to_cache(frame, video_id, frame_index)
            finally:
                if tmp_source is not None:
                    silent_remove(tmp_source)

    def add_image_to_cache(
        self, key: str, image: Union[np.ndarray, BinaryIO, bytes], ext=None
    ) -> np.ndarray:
        """
        Adds image to cache.
        """
        with self._lock:
            self._cache.save_image(key, image, ext)
            self._load_queue.delete(key)
            sly.logger.debug(f"Image {key} added to cache", extra={"image_id": key})
            return self._cache.get_image(key)

    def get_image_path(self, key) -> str:
        return str(self._cache.get_image_path(key))

    def get_video_path(self, key) -> str:
        return str(self._cache.get_video_path(key))

    def download_video(self, api: sly.Api, video_id: int, **kwargs) -> None:
        """
        Download video if needed and add it to cache. If video is already in cache, do nothing.
        If "return_images" in kwargs and is True, returns list of frames.
        """
        return_images = kwargs.get("return_images", True)
        progress_cb = kwargs.get("progress_cb", None)
        video_info = kwargs.get("video_info", None)
        if video_info is None:
            video_info = api.video.get_info_by_id(video_id)

        self._wait_if_in_queue(video_id, api.logger)
        if not video_id in self._cache:
            download_time = time.monotonic()
            self._load_queue.set(video_id, video_id)
            try:
                sly.logger.debug("Downloading video #%s", video_id)
                if progress_cb is None and self.log_progress:
                    size = video_info.file_meta.get("size", None)
                    if size is None:
                        size = "unknown"
                    else:
                        size = int(size)

                    prog_n = 0
                    prog_t = time.monotonic()

                    def _progress_cb(n):
                        nonlocal prog_n
                        nonlocal prog_t
                        prog_n += n
                        cur_t = time.monotonic()
                        if cur_t - prog_t > 3 or (isinstance(size, int) and prog_n >= size):
                            prog_t = cur_t
                            percent_str = ""
                            if isinstance(size, int):
                                percent_str = f" ({(prog_n*100) // size}%)"
                            prog_str = (
                                f"{(prog_n / 1000000):.2f}/{(size / 1000000):.2f} MB{percent_str}"
                            )
                            sly.logger.debug(
                                "Downloading video #%s: %s",
                                video_id,
                                prog_str,
                            )

                    progress_cb = _progress_cb
                temp_video_path = Path("/tmp/smart_cache").joinpath(
                    f"_{sly.rand_str(6)}_" + video_info.name
                )
                api.video.download_path(video_id, temp_video_path, progress_cb=progress_cb)
                self.add_video_to_cache(video_id, temp_video_path)
                download_time = time.monotonic() - download_time
                api.logger.debug(
                    f"Video #{video_id} downloaded to cache in {download_time:.2f} sec",
                    extra={"video_id": video_id, "download_time": download_time},
                )
            except Exception as e:
                self._load_queue.delete(video_id)
                raise e
        if return_images:
            return self.get_frames_from_cache(video_id, list(range(video_info.frames_count)))

    def add_cache_endpoint(self, server: FastAPI):
        @server.post("/smart_cache")
        def cache_endpoint(request: Request, task: BackgroundTasks):
            task.add_task(
                self.cache_task,
                api=request.state.api,
                state=request.state.state,
            )
            return {"message": "Cache task started."}

    def add_cache_files_endpoint(self, server: FastAPI):
        @server.post("/smart_cache_files")
        def cache_files_endpoint(
            request: Request,
            task: BackgroundTasks,
            files: List[UploadFile],
            settings: str = Form("{}"),
        ):
            state = json.loads(settings)
            task.add_task(
                self.cache_files_task,
                files=files,
                state=state,
            )
            return {"message": "Cache task started."}

    def cache_task(self, api: sly.Api, state: dict):
        if "server_address" in state and "api_token" in state:
            api = sly.Api(state["server_address"], state["api_token"])
        api.logger.debug("Request state in cache endpoint", extra=state)
        image_ids, task_type = self._parse_state(state)
        kwargs = {"return_images": False}

        if task_type is InferenceImageCache._LoadType.ImageId:
            if "dataset_id" in state:
                self.download_images(api, state["dataset_id"], image_ids, **kwargs)
            else:
                for img_id in image_ids:
                    self.download_image(api, img_id)
        elif task_type is InferenceImageCache._LoadType.ImageHash:
            self.download_images_by_hashes(api, image_ids, **kwargs)
        elif task_type is InferenceImageCache._LoadType.Frame:
            video_id = state["video_id"]
            self.download_frames(api, video_id, image_ids, **kwargs)
        elif task_type is InferenceImageCache._LoadType.Video:
            video_id = image_ids
            self.download_video(api, video_id, **kwargs)

    def add_frame_to_cache(self, frame: np.ndarray, video_id: int, frame_index: int):
        name = self._frame_name(video_id, frame_index)
        self._add_to_cache(name, frame)

    def cache_files_task(self, files: List[UploadFile], state: dict):
        sly.logger.debug("Request state in cache endpoint", extra=state)
        image_ids, task_type = self._parse_state(state)

        if task_type in (
            InferenceImageCache._LoadType.ImageId,
            InferenceImageCache._LoadType.ImageHash,
        ):
            for image_id, file in zip(image_ids, files):
                image = sly.image.read_bytes(file.file.read())
                name = self._image_name(image_id)
                self._add_to_cache(name, image)
        elif task_type is InferenceImageCache._LoadType.Frame:
            video_id = state["video_id"]
            for frame_index, file in zip(image_ids, files):
                frame = sly.image.read_bytes(file.file.read())
                self.add_frame_to_cache(frame, video_id, frame_index)
        elif task_type is InferenceImageCache._LoadType.Video:
            video_id = image_ids
            temp_video_path = Path("/tmp/smart_cache").joinpath(
                f"_{sly.rand_str(6)}_" + files[0].file.name
            )
            with open(temp_video_path, "wb") as f:
                shutil.copyfileobj(files[0].file, f)
            self._wait_if_in_queue(video_id, sly.logger)
            self._load_queue.set(video_id, video_id)
            self.add_video_to_cache(video_id, str(temp_video_path))

    def run_cache_task_manually(
        self,
        api: sly.Api,
        list_of_ids_ranges_or_hashes: List[Union[str, int, List[int]]],
        *,
        dataset_id: Optional[int] = None,
        video_id: Optional[int] = None,
    ) -> None:
        """
        Run cache_task in new thread.

        :param api: supervisely api
        :type api: sly.Api
        :param list_of_ids_ranges_or_hashes: information abount images/frames need to be loaded;
        to download images, pass list of integer IDs (`dataset_id` requires)
        or list of hash strings (`dataset_id` could be None);
        to download frames, pass list of pairs of indices of the first and last frame
        and `video_id` (ex.: [[1, 3], [5, 5], [7, 10]]);
        to download video, pass None and `video_id` (only for persistent cache)
        :type list_of_ids_ranges_or_hashes: List[Union[str, int, List[int]]]
        :param dataset_id: id of dataset on supervisely platform; default is None
        :type dataset_id: Optional[int]
        :param video_id: id of video on supervisely platform; default is None
        :type video_id: Optional[int]
        """
        state = {}
        if list_of_ids_ranges_or_hashes is None:
            api.logger.debug("Got a task to add video to cache")
            if not isinstance(self._cache, PersistentImageTTLCache):
                raise ValueError("Video can be added only to persistent cache")
            state["video_id"] = video_id
        elif isinstance(list_of_ids_ranges_or_hashes[0], str):
            api.logger.debug("Got a task to add images using hash")
            state["image_hashes"] = list_of_ids_ranges_or_hashes
        elif video_id is None:
            if dataset_id is None:
                api.logger.error("dataset_id or video_id must be defined if not hashes are used")
                return
            api.logger.debug("Got a task to add images using IDs")
            state["image_ids"] = list_of_ids_ranges_or_hashes
            state["dataset_id"] = dataset_id
        else:
            api.logger.debug("Got a task to add frames")
            state["video_id"] = video_id
            state["frame_ranges"] = list_of_ids_ranges_or_hashes

        self._download_executor.submit(self.cache_task, api=api, state=state)

    def set_project_meta(self, project_id, project_meta):
        pr_meta_name = self._project_meta_name(project_id)
        if isinstance(self._cache, PersistentImageTTLCache):
            self._cache.save_project_meta(pr_meta_name, project_meta)
        else:
            self._cache[pr_meta_name] = project_meta

    def get_project_meta(self, api: sly.Api, project_id: int):
        pr_meta_name = self._project_meta_name(project_id)
        if isinstance(self._cache, PersistentImageTTLCache):
            if pr_meta_name in self._cache:
                return self._cache.get_project_meta(pr_meta_name)
            project_meta = sly.ProjectMeta.from_json(api.project.get_meta(project_id))
            self._cache.save_project_meta(pr_meta_name, project_meta)
            return project_meta
        else:
            if pr_meta_name in self._cache:
                return self._cache[pr_meta_name]
            project_meta = sly.ProjectMeta.from_json(api.project.get_meta(project_id))
            self._cache[pr_meta_name] = project_meta
            return project_meta

    @property
    def ttl(self):
        return self._ttl

    @property
    def tmp_path(self):
        if self.is_persistent:
            return str(self._data_dir)
        return None

    def _parse_state(self, state: dict) -> Tuple[List[Any], _LoadType]:
        if "image_ids" in state:
            return state["image_ids"], InferenceImageCache._LoadType.ImageId
        elif "image_hashes" in state:
            return state["image_hashes"], InferenceImageCache._LoadType.ImageHash
        elif "video_id" in state:
            if "frame_ranges" in state:
                frame_ranges = state["frame_ranges"]
                frames = []
                for fr_range in frame_ranges:
                    shift = 1
                    if fr_range[0] > fr_range[1]:
                        shift = -1
                    start, end = fr_range[0], fr_range[1] + shift
                    frames.extend(list(range(start, end, shift)))
                return frames, InferenceImageCache._LoadType.Frame
            elif "frame_indexes" in state:
                return state["frame_indexes"], InferenceImageCache._LoadType.Frame
            return state["video_id"], InferenceImageCache._LoadType.Video
        raise ValueError("State has no proper fields: image_ids, image_hashes or video_id")

    def _add_to_cache(
        self,
        names: Union[str, List[str]],
        images: Union[np.ndarray, List[np.ndarray]],
    ):
        if isinstance(names, str):
            names = [names]

        if isinstance(images, np.ndarray):
            images = [images]

        if len(images) != len(names):
            raise ValueError(
                f"Number of images and names do not match: {len(images)} != {len(names)}"
            )

        for name, img in zip(names, images):
            with self._lock:
                self._cache.save_image(name, img)
                self._load_queue.delete(name)

    def _image_name(self, id_or_hash: Union[str, int]) -> str:
        if isinstance(id_or_hash, int):
            return f"image_{id_or_hash}"
        hash_wo_slash = id_or_hash.replace("/", "-")
        return f"image_{hash_wo_slash}"

    def _frame_name(self, video_id: int, frame_index: int) -> str:
        return f"frame_{video_id}_{frame_index}"

    def _video_name(self, video_id: int, video_name: str) -> str:
        ext = Path(video_name).suffix
        name = f"video_{video_id}{ext}"
        return name

    def _project_meta_name(self, project_id: int) -> str:
        return f"project_meta_{project_id}"

    def _download_many(
        self,
        indexes: List[Union[int, str]],
        name_cunstructor: Callable[[int], str],
        load_generator: Callable[
            [List[int]],
            Generator[Tuple[Union[int, str], np.ndarray], None, None],
        ],
        logger: Logger,
        return_images: bool = True,
        progress_cb=None,
        video_id=None,
    ) -> Optional[List[np.ndarray]]:
        pos_by_name = {}
        all_frames = [None for _ in range(len(indexes))]

        def get_one_image(item):
            pos, index = item
            if video_id in self._cache:
                return pos, self.get_frame_from_cache(video_id, index)
            return pos, self._cache.get_image(name_cunstructor(index))

        position = 0
        batch_size = 4
        for batch in batched(indexes, batch_size):
            indexes_to_load = []
            items = []
            for hash_or_id in batch:
                name = name_cunstructor(hash_or_id)
                self._wait_if_in_queue(name, logger)

                if name not in self._cache and video_id not in self._cache:
                    self._load_queue.set(name, hash_or_id)
                    indexes_to_load.append(hash_or_id)
                    pos_by_name[name] = position
                elif return_images is True:
                    items.append((position, hash_or_id))
                position += 1

            if len(items) > 0:
                with ThreadPoolExecutor(min(64, len(items))) as executor:
                    for pos, image in executor.map(get_one_image, items):
                        all_frames[pos] = image
                        if progress_cb is not None:
                            progress_cb()

            download_time = time.monotonic()
            if len(indexes_to_load) > 0:
                for id_or_hash, image in load_generator(indexes_to_load):
                    name = name_cunstructor(id_or_hash)
                    self._add_to_cache(name, image)

                    if return_images:
                        pos = pos_by_name[name]
                        all_frames[pos] = image
                        if progress_cb is not None:
                            progress_cb()
            download_time = time.monotonic() - download_time

            # logger.debug(f"All stored files: {sorted(os.listdir(self.tmp_path))}")
            if indexes_to_load:
                indexes_to_load = list(indexes_to_load)
                logger.debug(
                    f"Images/Frames added to cache: {indexes_to_load} in {download_time:.2f} sec",
                    extra={"indexes": indexes_to_load, "download_time": download_time},
                )
            found = set(batch).difference(indexes_to_load)
            if found:
                logger.debug(f"Images/Frames found in cache: {list(found)}")

        if return_images:
            return all_frames
        return

    def _wait_if_in_queue(self, name, logger: Logger):
        if name in self._load_queue:
            logger.debug(f"Waiting for other task to load {name}")

        while name in self._load_queue:
            # TODO: time.sleep if slowdown
            time.sleep(0.1)
            continue

    def download_frames_to_paths(self, api, video_id, frame_indexes, paths, progress_cb=None):
        def _download_frame(frame_index):
            self.download_frame(api, video_id, frame_index)
            name = self._frame_name(video_id, frame_index)
            return frame_index, name

        def _download_and_save(this_frame_indexes, this_paths):
            if video_id in self._cache:
                for path, frame in zip(
                    this_paths,
                    self._read_frames_from_cached_video_iter(video_id, this_frame_indexes),
                ):
                    sly.image.write(path, frame)
                    if progress_cb is not None:
                        progress_cb()
                return

            futures = []
            frame_index_to_path = {}
            for frame_index, path in zip(this_frame_indexes[:5], this_paths[:5]):
                frame_index_to_path[frame_index] = path
                futures.append(self._download_executor.submit(_download_frame, frame_index))
            for future in as_completed(futures):
                frame_index, name = future.result()
                path = frame_index_to_path[frame_index]
                self._cache.copy_to(name, path)
                if progress_cb is not None:
                    progress_cb()
            if len(this_frame_indexes) > 5:
                _download_and_save(this_frame_indexes[5:], this_paths[5:])

        # optimization for frame read from video file
        frame_indexes, paths = zip(*sorted(zip(frame_indexes, paths), key=lambda x: x[0]))
        _download_and_save(frame_indexes, paths)



from supervisely.app import DataJson
from supervisely.app.widgets import Widget
from typing import Dict, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class Markdown(Widget):
    def __init__(
        self,
        content: str = "",
        height: Union[int, Literal["fit-content"]] = "fit-content",
        widget_id: str = None,
        show_border: bool = True,
    ):
        self._content = content
        self._height = f"{height}px" if type(height) == int else height
        self._show_border = show_border

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        return {
            "content": self._content,
            "options": {"height": self._height},
            "showBorder": self._show_border,
        }

    def get_json_state(self) -> Dict:
        return {}

    def set_content(self, content: str) -> None:
        if not isinstance(content, str):
            raise TypeError("Content type has to be str.")
        self._content = content
        DataJson()[self.widget_id]["content"] = self._content
        DataJson().send_changes()

    def get_content(self) -> str:
        self._content = DataJson()[self.widget_id]["content"]
        return self._content

    def get_height(self) -> Union[int, Literal["fit-content"]]:
        self._height = DataJson()[self.widget_id]["options"]["height"]
        if self._height == "fit-content":
            return self._height
        return int(self._height[:-2])

    def set_height(self, height: Union[int, Literal["fit-content"]]) -> None:
        if type(height) != int and height != "fit-content":
            raise TypeError("Height value type has to be an integer or 'fit-content' string.")
        self._height = f"{height}px" if type(height) == int else height
        DataJson()[self.widget_id]["options"]["height"] = self._height
        DataJson().send_changes()



# coding: utf-8
"""api for working with tasks"""

import json
import os
import time
from collections import OrderedDict, defaultdict
from pathlib import Path

# docs
from typing import Any, Callable, Dict, List, Literal, NamedTuple, Optional, Union

import requests
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from tqdm import tqdm

from supervisely import logger
from supervisely._utils import batched, take_with_default
from supervisely.api.module_api import (
    ApiField,
    ModuleApiBase,
    ModuleWithStatus,
    WaitingTimeExceeded,
)
from supervisely.collection.str_enum import StrEnum
from supervisely.io.fs import (
    ensure_base_path,
    get_file_hash,
    get_file_name,
    get_file_name_with_ext,
)


class TaskFinishedWithError(Exception):
    """TaskFinishedWithError"""

    pass


class TaskApi(ModuleApiBase, ModuleWithStatus):
    """
    API for working with Tasks. :class:`TaskApi<TaskApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervisely.com", token="4r47N...xaTatb")

        task_id = 121230
        task_info = api.task.get_info_by_id(task_id)
    """

    class RestartPolicy(StrEnum):
        """RestartPolicy"""

        NEVER = "never"
        """"""
        ON_ERROR = "on_error"
        """"""

    class PluginTaskType(StrEnum):
        """PluginTaskType"""

        TRAIN = "train"
        """"""
        INFERENCE = "inference"
        """"""
        INFERENCE_RPC = "inference_rpc"
        """"""
        SMART_TOOL = "smarttool"
        """"""
        CUSTOM = "custom"
        """"""

    class Status(StrEnum):
        """Status"""

        QUEUED = "queued"
        """Application is queued for execution"""
        CONSUMED = "consumed"
        """Application is consumed by an agent"""
        STARTED = "started"
        """Application has been started"""
        DEPLOYED = "deployed"
        """Only for Plugins"""
        ERROR = "error"
        """Application has finished with an error"""
        FINISHED = "finished"
        """Application has finished successfully"""
        TERMINATING = "terminating"
        """Application is being terminated"""
        STOPPED = "stopped"
        """Application has been stopped"""

    def __init__(self, api):
        ModuleApiBase.__init__(self, api)
        ModuleWithStatus.__init__(self)

    def get_list(
        self, workspace_id: int, filters: Optional[List[Dict[str, str]]] = None
    ) -> List[NamedTuple]:
        """
        List of Tasks in the given Workspace.

        :param workspace_id: Workspace ID.
        :type workspace_id: int
        :param filters: List of params to sort output Projects.
        :type filters: List[dict], optional
        :return: List of Tasks with information for the given Workspace.
        :rtype: :class:`List[NamedTuple]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            workspace_id = 23821

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            task_infos = api.task.get_list(workspace_id)

            task_infos_filter = api.task.get_list(23821, filters=[{'field': 'id', 'operator': '=', 'value': 121230}])
            print(task_infos_filter)
            # Output: [
            #     {
            #         "id": 121230,
            #         "type": "clone",
            #         "status": "finished",
            #         "startedAt": "2019-12-19T12:13:09.702Z",
            #         "finishedAt": "2019-12-19T12:13:09.701Z",
            #         "meta": {
            #             "input": {
            #                 "model": {
            #                     "id": 1849
            #                 },
            #                 "isExternal": true,
            #                 "pluginVersionId": 84479
            #             },
            #             "output": {
            #                 "model": {
            #                     "id": 12380
            #                 },
            #                 "pluginVersionId": 84479
            #             }
            #         },
            #         "description": ""
            #     }
            # ]
        """
        return self.get_list_all_pages(
            "tasks.list",
            {ApiField.WORKSPACE_ID: workspace_id, ApiField.FILTER: filters or []},
        )

    def get_info_by_id(self, id: int) -> NamedTuple:
        """
        Get Task information by ID.

        :param id: Task ID in Supervisely.
        :type id: int
        :return: Information about Task.
        :rtype: :class:`NamedTuple`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            task_id = 121230

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            task_info = api.task.get_info_by_id(task_id)
            print(task_info)
            # Output: {
            #     "id": 121230,
            #     "workspaceId": 23821,
            #     "description": "",
            #     "type": "clone",
            #     "status": "finished",
            #     "startedAt": "2019-12-19T12:13:09.702Z",
            #     "finishedAt": "2019-12-19T12:13:09.701Z",
            #     "userId": 16154,
            #     "meta": {
            #         "app": {
            #             "id": 10370,
            #             "name": "Auto Import",
            #             "version": "test-branch",
            #             "isBranch": true,
            #         },
            #         "input": {
            #             "model": {
            #                 "id": 1849
            #             },
            #             "isExternal": true,
            #             "pluginVersionId": 84479
            #         },
            #         "output": {
            #             "model": {
            #                 "id": 12380
            #             },
            #             "pluginVersionId": 84479
            #         }
            #     },
            #     "settings": {},
            #     "agentName": null,
            #     "userLogin": "alexxx",
            #     "teamId": 16087,
            #     "agentId": null
            # }
        """
        return self._get_info_by_id(id, "tasks.info")

    def get_status(self, task_id: int) -> Status:
        """
        Check status of Task by ID.

        :param id: Task ID in Supervisely.
        :type id: int
        :return: Status object
        :rtype: :class:`Status`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            task_id = 121230

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            task_status = api.task.get_status(task_id)
            print(task_status)
            # Output: finished
        """
        status_str = self.get_info_by_id(task_id)[ApiField.STATUS]  # @TODO: convert json to tuple
        return self.Status(status_str)

    def raise_for_status(self, status: Status) -> None:
        """
        Raise error if Task status is ERROR.

        :param status: Status object.
        :type status: Status
        :return: None
        :rtype: :class:`NoneType`
        """
        if status is self.Status.ERROR:
            raise TaskFinishedWithError(f"Task finished with status {str(self.Status.ERROR)}")

    def wait(
        self,
        id: int,
        target_status: Status,
        wait_attempts: Optional[int] = None,
        wait_attempt_timeout_sec: Optional[int] = None,
    ):
        """
        Awaiting achievement by given Task of a given status.

        :param id: Task ID in Supervisely.
        :type id: int
        :param target_status: Status object(status of task we expect to destinate).
        :type target_status: Status
        :param wait_attempts: The number of attempts to determine the status of the task that we are waiting for.
        :type wait_attempts: int, optional
        :param wait_attempt_timeout_sec: Number of seconds for intervals between attempts(raise error if waiting time exceeded).
        :type wait_attempt_timeout_sec: int, optional
        :return: True if the desired status is reached, False otherwise
        :rtype: :class:`bool`
        """
        wait_attempts = wait_attempts or self.MAX_WAIT_ATTEMPTS
        effective_wait_timeout = wait_attempt_timeout_sec or self.WAIT_ATTEMPT_TIMEOUT_SEC
        for attempt in range(wait_attempts):
            status = self.get_status(id)
            self.raise_for_status(status)
            if status in [
                target_status,
                self.Status.FINISHED,
                self.Status.DEPLOYED,
                self.Status.STOPPED,
            ]:
                return
            time.sleep(effective_wait_timeout)
        raise WaitingTimeExceeded(
            f"Waiting time exceeded: total waiting time {wait_attempts * effective_wait_timeout} seconds, i.e. {wait_attempts} attempts for {effective_wait_timeout} seconds each"
        )

    def get_context(self, id: int) -> Dict:
        """
        Get context information by task ID.

        :param id: Task ID in Supervisely.
        :type id: int
        :return: Context information in dict format
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            task_id = 121230

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            context = api.task.get_context(task_id)
            print(context)
            # Output: {
            #     "team": {
            #         "id": 16087,
            #         "name": "alexxx"
            #     },
            #     "workspace": {
            #         "id": 23821,
            #         "name": "my_super_workspace"
            #     }
            # }
        """
        response = self._api.post("GetTaskContext", {ApiField.ID: id})
        return response.json()

    def _convert_json_info(self, info: dict):
        """_convert_json_info"""
        return info

    def start(
        self,
        agent_id,
        app_id: Optional[int] = None,
        workspace_id: Optional[int] = None,
        description: Optional[str] = "application description",
        params: Dict[str, Any] = None,
        log_level: Optional[Literal["info", "debug", "warning", "error"]] = "info",
        users_ids: Optional[List[int]] = None,
        app_version: Optional[str] = "",
        is_branch: Optional[bool] = False,
        task_name: Optional[str] = "pythonSpawned",
        restart_policy: Optional[Literal["never", "on_error"]] = "never",
        proxy_keep_url: Optional[bool] = False,
        module_id: Optional[int] = None,
        redirect_requests: Optional[Dict[str, int]] = {},
        limit_by_workspace: bool = False,
    ) -> Dict[str, Any]:
        """Starts the application task on the agent.

        :param agent_id: Agent ID. Can be obtained from TeamCluster page in UI.
        :type agent_id: int
        :param app_id: Deprecated. Use module_id instead.
        :type app_id: int, optional
        :param workspace_id: Workspace ID where the task will be created.
        :type workspace_id: int, optional
        :param description: Task description which will be shown in UI.
        :type description: str, optional
        :param params: Task parameters which will be passed to the application, check the
            code example below for more details.
        :type params: Dict[str, Any], optional
        :param log_level: Log level for the application.
        :type log_level: Literal["info", "debug", "warning", "error"], optional
        :param users_ids: List of user IDs for which will be created an instance of the application.
            For each user a separate task will be created.
        :type users_ids: List[int], optional
        :param app_version: Application version e.g. "v1.0.0" or branch name e.g. "dev".
        :type app_version: str, optional
        :param is_branch: If the application version is a branch name, set this parameter to True.
        :type is_branch: bool, optional
        :param task_name: Task name which will be shown in UI.
        :type task_name: str, optional
        :param restart_policy: when the task should be restarted: never or if error occurred.
        :type restart_policy: Literal["never", "on_error"], optional
        :param proxy_keep_url: For internal usage only.
        :type proxy_keep_url: bool, optional
        :param module_id: Module ID. Can be obtained from the apps page in UI.
        :type module_id: int, optional
        :param redirect_requests: For internal usage only in Develop and Debug mode.
        :type redirect_requests: Dict[str, int], optional
        :param limit_by_workspace: If set to True tasks will be only visible inside of the workspace
            with specified workspace_id.
        :type limit_by_workspace: bool, optional
        :return: Task information in JSON format.
        :rtype: Dict[str, Any]

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            app_slug = "supervisely-ecosystem/export-to-supervisely-format"
            module_id = api.app.get_ecosystem_module_id(app_slug)
            module_info = api.app.get_ecosystem_module_info(module_id)

            project_id = 12345
            agent_id = 12345
            workspace_id = 12345

            params = module_info.get_arguments(images_project=project_id)

            session = api.app.start(
                agent_id=agent_id,
                module_id=module_id,
                workspace_id=workspace_id,
                task_name="Prepare download link",
                params=params,
                app_version="dninja",
                is_branch=True,
            )
        """
        if app_id is not None and module_id is not None:
            raise ValueError("Only one of the arguments (app_id or module_id) have to be defined")
        if app_id is None and module_id is None:
            raise ValueError("One of the arguments (app_id or module_id) have to be defined")

        advanced_settings = {
            ApiField.LIMIT_BY_WORKSPACE: limit_by_workspace,
        }

        data = {
            ApiField.AGENT_ID: agent_id,
            # "nodeId": agent_id,
            ApiField.WORKSPACE_ID: workspace_id,
            ApiField.DESCRIPTION: description,
            ApiField.PARAMS: take_with_default(params, {"state": {}}),
            ApiField.LOG_LEVEL: log_level,
            ApiField.USERS_IDS: take_with_default(users_ids, []),
            ApiField.APP_VERSION: app_version,
            ApiField.IS_BRANCH: is_branch,
            ApiField.TASK_NAME: task_name,
            ApiField.RESTART_POLICY: restart_policy,
            ApiField.PROXY_KEEP_URL: proxy_keep_url,
            ApiField.ADVANCED_SETTINGS: advanced_settings,
        }
        if len(redirect_requests) > 0:
            data[ApiField.REDIRECT_REQUESTS] = redirect_requests

        if app_id is not None:
            data[ApiField.APP_ID] = app_id
        if module_id is not None:
            data[ApiField.MODULE_ID] = module_id
        resp = self._api.post(method="tasks.run.app", data=data)
        task = resp.json()[0]
        if "id" not in task:
            task["id"] = task.get("taskId")
        return task

    def stop(self, id: int):
        """stop"""
        response = self._api.post("tasks.stop", {ApiField.ID: id})
        return self.Status(response.json()[ApiField.STATUS])

    def submit_logs(self, logs) -> None:
        """submit_logs"""
        response = self._api.post("tasks.logs.add", {ApiField.LOGS: logs})
        # return response.json()[ApiField.TASK_ID]

    def upload_files(
        self,
        task_id: int,
        abs_paths: List[str],
        names: List[str],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> None:
        """upload_files"""
        if len(abs_paths) != len(names):
            raise RuntimeError("Inconsistency: len(abs_paths) != len(names)")

        hashes = []
        if len(abs_paths) == 0:
            return

        hash_to_items = defaultdict(list)
        hash_to_name = defaultdict(list)
        for idx, item in enumerate(zip(abs_paths, names)):
            path, name = item
            item_hash = get_file_hash(path)
            hashes.append(item_hash)
            hash_to_items[item_hash].append(path)
            hash_to_name[item_hash].append(name)

        unique_hashes = set(hashes)
        remote_hashes = self._api.image.check_existing_hashes(list(unique_hashes))
        new_hashes = unique_hashes - set(remote_hashes)

        # @TODO: upload remote hashes
        if len(remote_hashes) != 0:
            files = []
            for hash in remote_hashes:
                for name in hash_to_name[hash]:
                    files.append({ApiField.NAME: name, ApiField.HASH: hash})
            for batch in batched(files):
                resp = self._api.post(
                    "tasks.files.bulk.add-by-hash",
                    {ApiField.TASK_ID: task_id, ApiField.FILES: batch},
                )
        if progress_cb is not None:
            progress_cb(len(remote_hashes))

        for batch in batched(list(zip(abs_paths, names, hashes))):
            content_dict = OrderedDict()
            for idx, item in enumerate(batch):
                path, name, hash = item
                if hash in remote_hashes:
                    continue
                content_dict["{}".format(idx)] = json.dumps({"fullpath": name, "hash": hash})
                content_dict["{}-file".format(idx)] = (name, open(path, "rb"), "")

            if len(content_dict) > 0:
                encoder = MultipartEncoder(fields=content_dict)
                resp = self._api.post("tasks.files.bulk.upload", encoder)
                if progress_cb is not None:
                    progress_cb(len(content_dict))

    # {
    #     data: {my_val: 1}
    #     obj: {val: 1, res: 2}
    # }
    # {
    #     obj: {new_val: 1}
    # }
    # // apped: true, recursive: false
    # {
    #     data: {my_val: 1}
    #     obj: {new_val: 1}
    # }(edited)
    # // append: false, recursive: false
    # {
    #     obj: {new_val: 1}
    # }(edited)
    #
    # 16: 32
    # // append: true, recursive: true
    # {
    #     data: {my_val: 1}
    #     obj: {val: 1, res: 2, new_val: 1}
    # }

    def set_fields(self, task_id: int, fields: List) -> Dict:
        """set_fields"""
        for idx, obj in enumerate(fields):
            for key in [ApiField.FIELD, ApiField.PAYLOAD]:
                if key not in obj:
                    raise KeyError("Object #{} does not have field {!r}".format(idx, key))
        data = {ApiField.TASK_ID: task_id, ApiField.FIELDS: fields}
        resp = self._api.post("tasks.data.set", data)
        return resp.json()

    def set_fields_from_dict(self, task_id: int, d: Dict) -> Dict:
        """set_fields_from_dict"""
        fields = []
        for k, v in d.items():
            fields.append({ApiField.FIELD: k, ApiField.PAYLOAD: v})
        return self.set_fields(task_id, fields)

    def set_field(
        self,
        task_id: int,
        field: Dict,
        payload: Dict,
        append: Optional[bool] = False,
        recursive: Optional[bool] = False,
    ) -> Dict:
        """set_field"""
        fields = [
            {
                ApiField.FIELD: field,
                ApiField.PAYLOAD: payload,
                ApiField.APPEND: append,
                ApiField.RECURSIVE: recursive,
            }
        ]
        return self.set_fields(task_id, fields)

    def get_fields(self, task_id, fields: List):
        """get_fields"""
        data = {ApiField.TASK_ID: task_id, ApiField.FIELDS: fields}
        resp = self._api.post("tasks.data.get", data)
        return resp.json()["result"]

    def get_field(self, task_id: int, field: str):
        """get_field"""
        result = self.get_fields(task_id, [field])
        return result[field]

    def _set_output(self):
        """_set_output"""
        pass

    def set_output_project(
        self,
        task_id: int,
        project_id: int,
        project_name: Optional[str] = None,
        project_preview: Optional[str] = None,
    ) -> Dict:
        """set_output_project"""
        if project_name is None:
            project = self._api.project.get_info_by_id(project_id, raise_error=True)
            project_name = project.name
            project_preview = project.image_preview_url

        output = {ApiField.PROJECT: {ApiField.ID: project_id, ApiField.TITLE: project_name}}
        if project_preview is not None:
            output[ApiField.PROJECT][ApiField.PREVIEW] = project_preview
        resp = self._api.post(
            "tasks.output.set", {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output}
        )
        return resp.json()

    def set_output_report(
        self,
        task_id: int,
        file_id: int,
        file_name: str,
        description: Optional[str] = "Report",
    ) -> Dict:
        """set_output_report"""
        return self._set_custom_output(
            task_id,
            file_id,
            file_name,
            description=description,
            icon="zmdi zmdi-receipt",
        )

    def _set_custom_output(
        self,
        task_id,
        file_id,
        file_name,
        file_url=None,
        description="File",
        icon="zmdi zmdi-file-text",
        color="#33c94c",
        background_color="#d9f7e4",
        download=False,
    ):
        """_set_custom_output"""
        if file_url is None:
            file_url = self._api.file.get_url(file_id)

        output = {
            ApiField.GENERAL: {
                "icon": {
                    "className": icon,
                    "color": color,
                    "backgroundColor": background_color,
                },
                "title": file_name,
                "titleUrl": file_url,
                "download": download,
                "description": description,
            }
        }
        resp = self._api.post(
            "tasks.output.set", {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output}
        )
        return resp.json()

    def set_output_archive(
        self, task_id: int, file_id: int, file_name: str, file_url: Optional[str] = None
    ) -> Dict:
        """set_output_archive"""
        if file_url is None:
            file_url = self._api.file.get_info_by_id(file_id).storage_path
        return self._set_custom_output(
            task_id,
            file_id,
            file_name,
            file_url=file_url,
            description="Download archive",
            icon="zmdi zmdi-archive",
            download=True,
        )

    def set_output_file_download(
        self,
        task_id: int,
        file_id: int,
        file_name: str,
        file_url: Optional[str] = None,
        download: Optional[bool] = True,
    ) -> Dict:
        """set_output_file_download"""
        if file_url is None:
            file_url = self._api.file.get_info_by_id(file_id).storage_path
        return self._set_custom_output(
            task_id,
            file_id,
            file_name,
            file_url=file_url,
            description="Download file",
            icon="zmdi zmdi-file",
            download=download,
        )

    def send_request(
        self,
        task_id: int,
        method: str,
        data: Dict,
        context: Optional[Dict] = {},
        skip_response: bool = False,
        timeout: Optional[int] = 60,
        outside_request: bool = True,
        retries: int = 10,
        raise_error: bool = False,
    ):
        """send_request"""
        if type(data) is not dict:
            raise TypeError("data argument has to be a dict")
        context["outside_request"] = outside_request
        resp = self._api.post(
            "tasks.request.direct",
            {
                ApiField.TASK_ID: task_id,
                ApiField.COMMAND: method,
                ApiField.CONTEXT: context,
                ApiField.STATE: data,
                "skipResponse": skip_response,
                "timeout": timeout,
            },
            retries=retries,
            raise_error=raise_error,
        )
        return resp.json()

    def set_output_directory(self, task_id, file_id, directory_path):
        """set_output_directory"""
        return self._set_custom_output(
            task_id,
            file_id,
            directory_path,
            description="Directory",
            icon="zmdi zmdi-folder",
        )

    def update_meta(
        self,
        id: int,
        data: dict,
        agent_storage_folder: str = None,
        relative_app_dir: str = None,
    ):
        """
        Update given task metadata
        :param id: int — task id
        :param data: dict — meta data to update
        """
        if type(data) == dict:
            data.update({"id": id})
            if agent_storage_folder is None and relative_app_dir is not None:
                raise ValueError(
                    "Both arguments (agent_storage_folder and relative_app_dir) has to be defined or None"
                )
            if agent_storage_folder is not None and relative_app_dir is None:
                raise ValueError(
                    "Both arguments (agent_storage_folder and relative_app_dir) has to be defined or None"
                )
            if agent_storage_folder is not None and relative_app_dir is not None:
                data["agentStorageFolder"] = {
                    "hostDir": agent_storage_folder,
                    "folder": relative_app_dir,
                }

        self._api.post("tasks.meta.update", data)

    def _update_app_content(self, task_id: int, data_patch: List[Dict] = None, state: Dict = None):
        payload = {}
        if data_patch is not None and len(data_patch) > 0:
            payload[ApiField.DATA] = data_patch
        if state is not None and len(state) > 0:
            payload[ApiField.STATE] = state

        resp = self._api.post(
            "tasks.app-v2.data.set",
            {ApiField.TASK_ID: task_id, ApiField.PAYLOAD: payload},
        )
        return resp.json()

    def set_output_error(
        self,
        task_id: int,
        title: str,
        description: Optional[str] = None,
        show_logs: Optional[bool] = True,
    ) -> Dict:
        """
        Set custom error message to the task output.

        :param task_id: Application task ID.
        :type task_id: int
        :param title: Error message to be displayed in the task output.
        :type title: str
        :param description: Description to be displayed in the task output.
        :type description: Optional[str]
        :param show_logs: If True, the link to the task logs will be displayed in the task output.
        :type show_logs: Optional[bool], default True
        :return: Response JSON.
        :rtype: Dict
        :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            if sly.is_development():
               load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            task_id = 12345
            title = "Something went wrong"
            description = "Please check the task logs"
            show_logs = True
            api.task.set_output_error(task_id, title, description, show_logs)
        """

        output = {
            ApiField.GENERAL: {
                "icon": {
                    "className": "zmdi zmdi-alert-octagon",
                    "color": "#ff83a6",
                    "backgroundColor": "#ffeae9",
                },
                "title": title,
                "showLogs": show_logs,
                "isError": True,
            }
        }

        if description is not None:
            output[ApiField.GENERAL]["description"] = description

        resp = self._api.post(
            "tasks.output.set",
            {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output},
        )
        return resp.json()

    def set_output_text(
        self,
        task_id: int,
        title: str,
        description: Optional[str] = None,
        show_logs: Optional[bool] = False,
        zmdi_icon: Optional[str] = "zmdi-comment-alt-text",
        icon_color: Optional[str] = "#33c94c",
        background_color: Optional[str] = "#d9f7e4",
    ) -> Dict:
        """
        Set custom text message to the task output.

        :param task_id: Application task ID.
        :type task_id: int
        :param title: Text message to be displayed in the task output.
        :type title: str
        :param description: Description to be displayed in the task output.
        :type description: Optional[str]
        :param show_logs: If True, the link to the task logs will be displayed in the task output.
        :type show_logs: Optional[bool], default False
        :param zmdi_icon: Icon class name from Material Design Icons (ZMDI).
        :type zmdi_icon: Optional[str], default "zmdi-comment-alt-text"
        :param icon_color: Icon color in HEX format.
        :type icon_color: Optional[str], default "#33c94c" (nearest Duron Jolly Green)
        :param background_color: Background color in HEX format.
        :type background_color: Optional[str], default "#d9f7e4" (Cosmic Latte)
        :return: Response JSON.
        :rtype: Dict
        :Usage example:

        .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            task_id = 12345
            title = "Task is finished"
            api.task.set_output_text(task_id, title)
        """

        output = {
            ApiField.GENERAL: {
                "icon": {
                    "className": f"zmdi {zmdi_icon}",
                    "color": icon_color,
                    "backgroundColor": background_color,
                },
                "title": title,
                "showLogs": show_logs,
                "isError": False,
            }
        }

        if description is not None:
            output[ApiField.GENERAL]["description"] = description

        resp = self._api.post(
            "tasks.output.set",
            {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output},
        )
        return resp.json()

    def update_status(
        self,
        task_id: int,
        status: Status,
    ) -> None:
        """Sets the specified status for the task.

        :param task_id: Task ID in Supervisely.
        :type task_id: int
        :param status: Task status to set.
        :type status: One of the values from :class:`Status`, e.g. Status.FINISHED, Status.ERROR, etc.
        :raises ValueError: If the status value is not allowed.
        """
        # If status was passed without converting to string, convert it.
        # E.g. Status.FINISHED -> "finished"
        status = str(status)
        if status not in self.Status.values():
            raise ValueError(
                f"Invalid status value: {status}. Allowed values: {self.Status.values()}"
            )
        self._api.post("tasks.status.update", {ApiField.ID: task_id, ApiField.STATUS: status})

    def set_output_experiment(self, task_id: int, experiment_info: dict) -> Dict:
        """
        Sets output for the task with experiment info.

        :param task_id: Task ID in Supervisely.
        :type task_id: int
        :param experiment_info: Experiment info from TrainApp.
        :type experiment_info: dict
        :return: None
        :rtype: :class:`NoneType`

        Example of experiment_info:

            experiment_info = {
                'experiment_name': '247_Lemons_RT-DETRv2-M',
                'framework_name': 'RT-DETRv2',
                'model_name': 'RT-DETRv2-M',
                'task_type': 'object detection',
                'project_id': 76,
                'task_id': 247,
                'model_files': {'config': 'model_config.yml'},
                'checkpoints': ['checkpoints/best.pth', 'checkpoints/checkpoint0025.pth', 'checkpoints/checkpoint0050.pth', 'checkpoints/last.pth'],
                'best_checkpoint': 'best.pth',
                'export': {'ONNXRuntime': 'export/best.onnx'},
                'app_state': 'app_state.json',
                'model_meta': 'model_meta.json',
                'train_val_split': 'train_val_split.json',
                'train_size': 4,
                'val_size': 2,
                'hyperparameters': 'hyperparameters.yaml',
                'hyperparameters_id': 45234,
                'artifacts_dir': '/experiments/76_Lemons/247_RT-DETRv2/',
                'datetime': '2025-01-22 18:13:43',
                'evaluation_report_id': 12961,
                'evaluation_report_link': 'https://app.supervisely.com/model-benchmark?id=12961',
                'evaluation_metrics': {
                    'mAP': 0.994059405940594,
                    'AP50': 1.0, 'AP75': 1.0,
                    'f1': 0.9944444444444445,
                    'precision': 0.9944444444444445,
                    'recall': 0.9944444444444445,
                    'iou': 0.9726227736959404,
                    'classification_accuracy': 1.0,
                    'calibration_score': 0.8935745942476048,
                    'f1_optimal_conf': 0.500377893447876,
                    'expected_calibration_error': 0.10642540575239527,
                    'maximum_calibration_error': 0.499622106552124
                },
                'primary_metric': 'mAP'
                'logs': {
                    'type': 'tensorboard',
                    'link': '/experiments/76_Lemons/247_RT-DETRv2/logs/'
                },
            }
        """
        output = {
            ApiField.EXPERIMENT: {ApiField.DATA: {**experiment_info}},
        }
        resp = self._api.post(
            "tasks.output.set", {ApiField.TASK_ID: task_id, ApiField.OUTPUT: output}
        )
        return resp.json()

    def is_running(self, task_id: int) -> bool:
        """
        Check if the task is running.

        :param task_id: Task ID in Supervisely.
        :type task_id: int
        :return: True if the task is running, False otherwise.
        :rtype: bool
        """
        try:
            self.send_request(task_id, "is_running", {}, retries=1, raise_error=True)
        except requests.exceptions.HTTPError as e:
            return False
        return True

    def is_ready(self, task_id: int) -> bool:
        """
        Check if the task is ready.

        :param task_id: Task ID in Supervisely.
        :type task_id: int
        :return: True if the task is ready, False otherwise.
        :rtype: bool
        """
        try:
            return (
                self.send_request(task_id, "is_ready", {}, retries=1, raise_error=True)["status"]
                == "ready"
            )
        except requests.exceptions.HTTPError as e:
            return False



import cv2
import numpy as np


def one_hot(segmentation, num_classes):
    # TODO: this assumes that the classes start with 0
    segmentation_copy = segmentation.copy()
    eye = np.eye(num_classes + 1, dtype=np.bool_)
    one_hot_seg = eye[segmentation_copy].transpose(2, 0, 1)
    return one_hot_seg[:num_classes]


def single_one_hot(segmentation, cls):
    one_hot_seg = np.zeros_like(segmentation)
    one_hot_seg[segmentation == cls] = 1
    return one_hot_seg


def get_single_contiguous_segment(one_hot_segmentation):
    import scipy  # pylint: disable=import-error

    kernel = np.ones((3, 3), dtype=one_hot_segmentation.dtype)
    seg = scipy.ndimage.label(one_hot_segmentation, structure=kernel)[0]
    return [np.where(seg == l) for l in range(1, seg.max() + 1)]


def get_contiguous_segments(one_hot_segmentation):
    import scipy  # pylint: disable=import-error

    kernel = np.ones((3, 3), dtype=one_hot_segmentation.dtype)
    segments_tensor = np.stack(
        [scipy.ndimage.label(seg, structure=kernel)[0] for seg in one_hot_segmentation]
    )

    return {
        c: [np.where(seg == l) for l in range(1, seg.max() + 1)]
        for c, seg in enumerate(segments_tensor)
    }


def get_exact_kernel(width):
    x = np.arange(-width, width + 1)
    xx, yy = np.meshgrid(x, x)
    d = np.sqrt(xx**2 + yy**2)
    kernel = (np.round(d) <= width).astype(np.uint8)
    return kernel


def erode_mask(mask, width, implementation="fast"):
    if mask.ndim == 2:
        input_mask = mask.astype(np.uint8)
    elif mask.ndim == 3:
        input_mask = mask.astype(np.uint8).transpose(1, 2, 0)
    else:
        raise ValueError(f"Assume an array of shape (H,W) or (C,H,W), got {mask.shape}!")

    if implementation == "fast":
        kernel = np.ones((3, 3), dtype=np.uint8)
        eroded_mask = cv2.erode(input_mask, kernel, iterations=width)
    elif implementation == "exact":
        kernel = get_exact_kernel(width)
        eroded_mask = cv2.erode(input_mask, kernel, iterations=1)
    else:
        raise ValueError(
            f'Implementation has to be one of "exact" and "fast", received: {implementation}!'
        )

    if input_mask.ndim == 2:
        return eroded_mask
    else:  # input_mask.ndim == 3
        if mask.shape[0] == 1:  # in this case, the first axis gets removed by cv2
            eroded_mask = np.expand_dims(eroded_mask, -1)
        return eroded_mask.transpose(2, 0, 1)


def dilate_mask(mask, width, implementation="fast"):
    if mask.ndim == 2:
        input_mask = mask.astype(np.uint8)
    elif mask.ndim == 3:
        input_mask = mask.astype(np.uint8).transpose(1, 2, 0)
    else:
        raise ValueError(f"Assume an array of shape (H,W) or (C,H,W), got {mask.shape}!")

    if implementation == "fast":
        kernel = np.ones((3, 3), dtype=np.uint8)
        dilated_mask = cv2.dilate(input_mask, kernel, iterations=width)
    elif implementation == "exact":
        kernel = get_exact_kernel(width)
        dilated_mask = cv2.dilate(input_mask, kernel, iterations=1)
    else:
        raise ValueError(
            f'Implementation has to be one of "exact" and "fast", received: {implementation}!'
        )

    if input_mask.ndim == 2:
        return dilated_mask
    else:  # input_mask.ndim == 3
        if mask.shape[0] == 1:  # in this case, the first axis gets removed by cv2
            dilated_mask = np.expand_dims(dilated_mask, -1)
        return dilated_mask.transpose(2, 0, 1)


def get_interior_boundary(mask, width, implementation="fast"):
    eroded_mask = erode_mask(mask, width, implementation)
    boundary_mask = (mask - eroded_mask).astype(np.bool_)
    return boundary_mask


def get_exterior_boundary(mask, width, implementation="fast"):
    dilated_mask = dilate_mask(mask, width, implementation)
    boundary_mask = (dilated_mask - mask).astype(np.bool_)
    return boundary_mask



# coding: utf-8


# docs
from __future__ import annotations

from typing import Dict, List, Optional, Tuple

import cv2

from supervisely._utils import unwrap_if_numpy
from supervisely.geometry.constants import (
    CLASS_ID,
    CREATED_AT,
    ID,
    LABELER_LOGIN,
    UPDATED_AT,
)
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.image_rotator import ImageRotator
from supervisely.geometry.point_location import PointLocation
from supervisely.geometry.rectangle import Rectangle


class Point(Geometry):
    """
    Point geometry for a single :class:`Label<supervisely.annotation.label.Label>`. :class:`Point<Point>` object is immutable.

    :param row: Position of Point on height.
    :type row: int or float
    :param col: Position of Point on width.
    :type col: int or float
    :param sly_id: Point ID in Supervisely server.
    :type sly_id: int, optional
    :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which Point belongs.
    :type class_id: int, optional
    :param labeler_login: Login of the user who created Point.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when Point was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when Point was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        row = 100
        col = 200
        figure = sly.Point(row, col)
    """

    def __init__(
        self,
        row: int,
        col: int,
        sly_id: Optional[int] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[int] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ):
        super().__init__(
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )
        self._row = round(unwrap_if_numpy(row))
        self._col = round(unwrap_if_numpy(col))

    @property
    def row(self) -> int:
        """
        Position of Point height.

        :return: Height of Point
        :rtype: :class:`int`
        :Usage example:

         .. code-block:: python

            print(figure.row)
            # Output: 100
        """
        return self._row

    @property
    def col(self) -> int:
        """
        Position of Point width.

        :return: Width of Point
        :rtype: :class:`int`
        :Usage example:

         .. code-block:: python

            print(figure.col)
            # Output: 200
        """
        return self._col

    @classmethod
    def from_point_location(
        cls,
        pt: PointLocation,
        sly_id: Optional[int] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[int] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ) -> Point:
        """
        Create Point from given :class:`PointLocation<supervisely.geometry.point_location.PointLocation>` object.

        :param pt: PointLocation object.
        :type pt: PointLocation
        :param sly_id: Point ID in Supervisely server.
        :type sly_id: int, optional
        :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which Point belongs.
        :type class_id: int, optional
        :param labeler_login: Login of the user who created Point.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when Point was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when Point was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional
        :return: Point object
        :rtype: :class:`Point<Point>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            figure_loc = sly.PointLocation(100, 200)
            figure = sly.Point.from_point_location(figure_loc)
        """
        return cls(
            row=pt.row,
            col=pt.col,
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    @property
    def point_location(self) -> PointLocation:
        """
        Create PointLocation object from Point.

        :return: PointLocation object
        :rtype: :class:`PointLocation<supervisely.geometry.point_location.PointLocation>`
        :Usage example:

         .. code-block:: python

            figure_loc = figure.point_location
        """
        return PointLocation(row=self.row, col=self.col)

    @staticmethod
    def geometry_name():
        """ """
        return "point"

    def crop(self, rect: Rectangle) -> List[Point]:
        """
        Crops current Point.

        :param rect: Rectangle object for crop.
        :type rect: Rectangle
        :return: List of Point objects
        :rtype: :class:`List[Point]<Point>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            crop_figures = figure.crop(sly.Rectangle(1, 1, 300, 350))
        """
        return [self.clone()] if rect.contains_point_location(self.point_location) else []

    def rotate(self, rotator: ImageRotator) -> Point:
        """
        Rotates current Point.

        :param rotator: ImageRotator object for rotation.
        :type rotator: ImageRotator
        :return: Point object
        :rtype: :class:`Point<Point>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.geometry.image_rotator import ImageRotator

            # Remember that Point class object is immutable, and we need to assign new instance of Point to a new variable
            height, width = 300, 400
            rotator = ImageRotator((height, width), 25)
            rotate_figure = figure.rotate(rotator)
        """
        return self.from_point_location(self.point_location.rotate(rotator))

    def resize(self, in_size: Tuple[int, int], out_size: Tuple[int, int]) -> Point:
        """
        Resizes current Point.

        :param in_size: Input Input image size (height, width) to which belongs Point.
        :type in_size: Tuple[int, int]
        :param out_size: Desired output image size (height, width) to which belongs Point.
        :type out_size: Tuple[int, int]
        :return: Point object
        :rtype: :class:`Point<Point>`

        :Usage Example:

         .. code-block:: python

            # Remember that Point class object is immutable, and we need to assign new instance of Point to a new variable
            in_height, in_width = 300, 400
            out_height, out_width = 600, 800
            resize_figure = figure.resize((in_height, in_width), (out_height, out_width))
        """
        return self.from_point_location(self.point_location.resize(in_size, out_size))

    def fliplr(self, img_size: Tuple[int, int]) -> Point:
        """
        Flips current Point in horizontal.

        :param img_size: Input image size (height, width) to which belongs Point object.
        :type img_size: Tuple[int, int]
        :return: Point object
        :rtype: :class:`Point<Point>`

        :Usage Example:

         .. code-block:: python

            # Remember that Point class object is immutable, and we need to assign new instance of Point to a new variable
            height, width = 300, 400
            fliplr_figure = figure.fliplr((height, width))
        """
        return self.from_point_location(self.point_location.fliplr(img_size))

    def flipud(self, img_size: Tuple[int, int]) -> Point:
        """
        Flips current Point in vertical.

        :param img_size: Input image size (height, width) to which belongs Point object.
        :type img_size: Tuple[int, int]
        :return: Point object
        :rtype: :class:`Point<Point>`

        :Usage Example:

         .. code-block:: python

            # Remember that Point class object is immutable, and we need to assign new instance of Point to a new variable
            height, width = 300, 400
            flipud_figure = figure.flipud((height, width))
        """
        return self.from_point_location(self.point_location.flipud(img_size))

    def scale(self, factor: float) -> Point:
        """
        Scales current Point.

        :param factor: Scale parameter.
        :type factor: float
        :return: Point object
        :rtype: :class:`Point<Point>`

        :Usage Example:

         .. code-block:: python

            # Remember that Point class object is immutable, and we need to assign new instance of Point to a new variable
            scale_figure = figure.scale(0.75)
        """
        return self.from_point_location(self.point_location.scale(factor))

    def translate(self, drow: int, dcol: int) -> Point:
        """
        Translates current Point.

        :param drow: Horizontal shift.
        :type drow: int
        :param dcol: Vertical shift.
        :type dcol: int
        :return: Point object
        :rtype: :class:`Point<Point>`

        :Usage Example:

         .. code-block:: python

            # Remember that Point class object is immutable, and we need to assign new instance of Point to a new variable
            translate_figure = figure.translate(150, 350)
        """
        return self.from_point_location(self.point_location.translate(drow, dcol))

    def _draw_impl(self, bitmap, color, thickness=1, config=None):
        r = round(thickness / 2)  # @TODO: relation between thickness and point radius - ???
        cv2.circle(bitmap, (self.col, self.row), radius=r, color=color, thickness=cv2.FILLED)

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        # @TODO: mb dummy operation for Point
        r = round((thickness + 1) / 2)
        cv2.circle(bitmap, (self.col, self.row), radius=r, color=color, thickness=cv2.FILLED)

    @property
    def area(self) -> float:
        """
        Point area.

        :return: Area of current Point object, always 0.0
        :rtype: :class:`float`

        :Usage Example:

         .. code-block:: python

            print(figure.area)
            # Output: 0.0
        """
        return 0.0

    def to_bbox(self) -> Rectangle:
        """
        Create Rectangle object from current Point.

        :return: Rectangle object
        :rtype: :class:`Rectangle<supervisely.geometry.rectangle.Rectangle>`

        :Usage Example:

         .. code-block:: python

            rectangle = figure.to_bbox()
        """
        return Rectangle(top=self.row, left=self.col, bottom=self.row, right=self.col)

    def to_json(self) -> Dict:
        """
        Convert the Point to a json dict. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :return: Json format as a dict
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            figure_json = figure.to_json()
            print(figure_json)
            # Output: {
            #    "points": {
            #        "exterior": [
            #            [200, 100]
            #        ],
            #        "interior": []
            #    }
            # }
        """
        res = self.point_location.to_json()
        self._add_creation_info(res)
        return res

    @classmethod
    def from_json(cls, data: Dict) -> Point:
        """
        Convert a json dict to Point. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: Point in json format as a dict.
        :type data: dict
        :return: Point object
        :rtype: :class:`Point<Point>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            figure_json = {
                "points": {
                    "exterior": [
                        [200, 100]
                    ],
                    "interior": []
                }
            }
            figure = sly.Point.from_json(figure_json)
        """
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls.from_point_location(
            PointLocation.from_json(data),
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    @classmethod
    def allowed_transforms(cls):
        """ """
        from supervisely.geometry.any_geometry import AnyGeometry

        return [AnyGeometry]
